--[[
Originally I was going to use LuaSocket/LuaSec to implement this but those
are kind of a PITA to use.  So, for the short term this is being supported via
use of 'curl'.
]]
local config = require 'system_config'
local cjson = require 'cjson.safe'
local uuid = require 'uuid'
local network_db = require 'network_db'

local server = 'hbmpushnotificationproxy.azurewebsites.net'

local senderid = config.device_type:find('edaqxr') and
    '80b64cf0-9732-43b1-8ced-8dfa9ced9227' or
    '8c22c42e-c06e-4524-a86e-b630e3bb22fe'

local cmd = {
    verify = [[/usr/bin/curl -k -X POST -H "Content-Type: application/json" -d '%s' https://%s/Api/PushChannelVerification 2>/dev/null]],
    newid = [[/usr/bin/curl -k -X POST -H "Content-Type: application/json" -d @%s https://%s/Api/PushRegistration 2>/dev/null]],
    push = [[/usr/bin/curl -k -X POST -H "Content-Type: application/json" -d @%s https://%s/Api/PushNotifications 2>/dev/null]]
}

local function send_request(fmt, json)
    local _cmd = string.format(fmt, json, server)
    local fh = io.popen(_cmd)
    if fh then
        for line in fh:lines() do
            local cj = cjson.decode(line)
            if cj then
                fh:close()
                return cj, cj.Message and true
            end
        end
        fh:close()
    end
end

local function _do_verify(chan_id)
    local cj = cjson.encode{
        senderid = senderid,
        pushchannel = chan_id
    }
    return send_request(cmd.verify, cj)
end

local function _do_newid(email_addr, description)
    local cj = cjson.encode{
        senderid = senderid,
        email = email_addr,
        description = description,
    }
    local filename = '/tmp/' .. uuid.generate() .. '.json'
    local fh = io.open(filename, 'w')
    if fh then
        fh:write(cj)
        fh:close()
        local r, e = send_request(cmd.newid, filename)
        os.remove(filename)
        return r, e
    end
end


local function _do_send(chan_id, subject, message, color)
    local ndb = network_db:new()
    local from = ndb:get_inside_network_hostname()
    local alias = ndb:get_alias()
    if alias then
        from = from .. " (" .. alias .. ")"
    end
    ndb:close()
    from = "From " .. from .. "\n"

    local cj = cjson.encode{
        senderid = senderid,
        subject = subject,
        text = from .. message,
        pushchannel = chan_id,
        color = color or '#00ff00',
    }
    local filename = '/tmp/' .. uuid.generate() .. '.json'
    local fh = io.open(filename, 'w')
    if fh then
        fh:write(cj)
        fh:close()
        local r, e = send_request(cmd.push, filename)
        os.remove(filename)
        return r, e
    end
end

local m = {
    verify = _do_verify,
    newid = _do_newid,
    send = _do_send
}

do return m end

--[===[
local t, err = m.verify('BKsqFL6SDs')
if err then print("ERROR", t.Message) end
--[[
local t, err = m.newid('mike.lynch@hbm.com', 'Test Event')
if err then print("ERROR", t.Message)
else print("Pending pushchannel", t.pushChannel)
end
]]
local t, err = m.send('BKsqFL6SDs', 'Test Stopped', 'The test has stopped because the unit has caught on fire', '#ff0000')
if err then print("ERROR", t.Message) end
]===]
