require 'strict'
require 'middleclass'
local ev = require 'ev'
local dbus = require 'dbus'
local cjson = require 'cjson.safe'

local bus = dbus.bus_get('system')
bus:hook_libev_default_loop()

local status_lib = class('status_lib')

function status_lib:initialize(module, callback)
    self.set_status = function()
        error('Status deamon services unavailable')
    end
    self.get_status = self.set_status

    self.status = bus:auto_proxy('com.somat.armada.Status',
        '/com/somat/armada/Status')
    if not self.status then
        error('Status daemon services unavailable')
    end

    self.module = module

    self.set_status = function(self, tbl)
        if type(tbl) ~= 'table' then
            error('Table expected, got ' .. type(tbl))
        end
        local ecj, err = cjson.encode(tbl)
        if not ecj then
            error('Unable to encode table to JSON: ' .. err)
        end
        self.status:UpdateStatus(self.module, ecj)
    end

    self.get_status = function(self, all)
        local st = cjson.decode(self.status:GetStatus())
        if all or not self.module then return st or {} end
        return st[self.module] or {}
    end
        
    if callback then
        self.callback = callback
        self.status.StatusChange:bind(callback)
    end
end

return status_lib
