local can_node2 = require 'test_engine/can_node2'
local gps_node = require 'test_engine/gps_node'
local digital_out_node = require 'test_engine/digital_out_node'
local _pn = require 'test_engine.pn_support'
local ffi = require 'ffi'
local cjson = require 'cjson.safe'
local config = require 'system_config'

local m = {}

local SIE_IDX_OFFSET = 3
local module_cache = {}

m.create_proxy_node = _pn.create_proxy_node
m.create_digital_out_node = _pn.create_digital_out_node
m.create_gps_signal_node = _pn.create_gps_signal_node
m.create_mx_can_node = _pn.create_mx_can_node
m.create_mx878_proxy_node = _pn.create_mx878_proxy_node

function m:init_rezero(node, stanza)
    -- If this is an SIE test and prerun balance mode is 'normal_test_starts'
    -- then set the rezero callback function for the node.
    if not stanza.purpose and stanza.prerun_balance and
        stanza.prerun_balance.mode and
        stanza.prerun_balance.mode == 'normal_test_starts' then
        self.rz_cb_connectors[stanza.name] = stanza.connector

        node.rezero_offset_cb = nil

        -- We are done if rezero offset exists
        if stanza.rezero_offset then return end

        -- If this is NOT a brg or hls channel set the callback
        if not stanza.connector:find("x*brg_") and
            not stanza.connector:find("x*hls_") then
            node.rezero_offset_cb = node.channel
        end
    end
end

function m:preset_rezero(stanza)
    if stanza.purpose then return end
    local connector = stanza.connector
    local fh = io.open(config.rezero_file_path .. connector, 'r')
    if not fh then return end
    local v = fh:read('*a')
    fh:close()
    if not v then return end
    return tonumber(v)
end

function m:initialize_raw_can(stanza)
    local pn, start, set_time = self:create_proxy_node(stanza)
    self.set_time = self.set_time or {}
    self.set_time[#self.set_time + 1] = set_time
    self.start_fn[#self.start_fn + 1] = {
        start = start,
        md = stanza,
    }
    self.proxy_nodes[stanza.connector] = pn
    self.all_nodes[stanza.name] = pn
    self.metas[stanza.name] = stanza
    if stanza.collect then
        pn.sie_idx = self.max_sie_buffers + SIE_IDX_OFFSET
        self.max_sie_buffers = self.max_sie_buffers + 1
    end
    self.raw_can_channels = self.raw_can_channels or {}
    self.raw_can_channels[stanza.connector] = pn
    self.can_connector_signal_count =
        self.can_connector_signal_count or {}
    self.sid_list[stanza.name] = self.stream_server:add_channel(stanza,
        pn.pipe_list)
    self.can_connector_signal_count[stanza.connector] = stanza.signal_count
    pn.can_set = can_node2.create_can_set()
end

function m:initialize_can_signal(stanza)
    local raw_can_connector = stanza.connector:match('^.+(@.+)$')
    local pn = self.proxy_nodes[raw_can_connector]
    stanza.rezero_offset = self:preset_rezero(stanza)
    local cs = can_node2.add_can_signal(pn.can_set, stanza,
        self.summary.pipe_rate or 5)
    self.all_nodes[stanza.name] = cs
    self.proxy_nodes[stanza.connector] = cs
    self.metas[stanza.name] = stanza
    if stanza.collect then
        cs.sie_idx = self.max_sie_buffers + SIE_IDX_OFFSET
        self.max_sie_buffers = self.max_sie_buffers + 1
    end
    stanza.data_type = string.gsub(stanza.data_type, 'message_', 'sequential_')
    self.sid_list[stanza.name] = self.stream_server:add_channel(stanza,
        cs.pipe_list)
    self.can_connector_signal_count[raw_can_connector] =
        self.can_connector_signal_count[raw_can_connector] - 1
    if self.can_connector_signal_count[raw_can_connector] == 0 then
        can_node2.finalize(pn.can_set)
    end
    self:init_rezero(cs, stanza)
end

function m:initialize_raw_gps(stanza)
    local pn, start, set_time = self:create_proxy_node(stanza)
    self.set_time = self.set_time or {}
    self.set_time[#self.set_time + 1] = set_time
    self.start_fn[#self.start_fn + 1] = {
        start = start,
        md = stanza,
    }
    self.proxy_nodes[stanza.connector] = pn
    self.all_nodes[stanza.name] = pn
    self.metas[stanza.name] = stanza
    if stanza.collect then
        pn.sie_idx = self.max_sie_buffers + SIE_IDX_OFFSET
        self.max_sie_buffers = self.max_sie_buffers + 1
    end

    self.sid_list[stanza.name] =
        self.stream_server:add_channel(stanza, pn.pipe_list)

    self.raw_gps_channels = self.raw_gps_channels or {}
    self.raw_gps_channels[stanza.connector] = pn
end

function m:initialize_gps_signal(stanza)
    stanza.rezero_offset = self:preset_rezero(stanza)
    local pn = self:create_gps_signal_node(stanza,
        self.raw_gps_channels[stanza.raw_channel],
        self.summary.pipe_rate)
    self.all_nodes[stanza.name] = pn
    self.proxy_nodes[stanza.connector] = pn
    self.metas[stanza.name] = stanza
    if stanza.data_type:find('_float') then
        stanza.data_type = string.gsub(stanza.data_type,
            'message_', 'sequential_')
    end
    self.sid_list[stanza.name] =
        self.stream_server:add_channel(stanza, pn.pipe_list)
    if stanza.collect then
        pn.sie_idx = self.max_sie_buffers + SIE_IDX_OFFSET
        self.max_sie_buffers = self.max_sie_buffers + 1
    end
    self:init_rezero(pn, stanza)
end

function m:initialize_mx471(stanza)
    -- If this is an MX471 CAN channel,
    -- do all sorts of special stuff
    self.mx_can_channels = self.mx_can_channels or {}
    local can_connector = stanza.connector:match('(@can.+)$')
    if not can_connector then
        error(string.format(
            'Unable to determine MX CAN connector from %s',
            stanza.connector))
    end
    if stanza.data_type:find('^message_') and stanza.sample_rate and
        tonumber(stanza.sample_rate) then
        stanza.data_type = string.gsub(stanza.data_type, 'message_',
            'sequential_')
    end
    local pn, start, cn, set_time
    if not self.mx_can_channels[can_connector] then
        stanza.rezero_offset = self:preset_rezero(stanza)
        pn, start, cn, set_time = self:create_mx_can_node(stanza, can_connector)
        self.start_fn[#self.start_fn + 1] = {
            start = start,
            md = stanza,
        }
        self.proxy_nodes[stanza.connector] = pn
        self.all_nodes[stanza.name] = cn
    else
        pn, start, cn, set_time =
            self:create_mx_can_node(stanza, can_connector)
        self.all_nodes[stanza.name] = cn
    end
    self.metas[stanza.name] = stanza
    if stanza.collect then
        cn.sie_idx = self.max_sie_buffers + SIE_IDX_OFFSET
        self.max_sie_buffers = self.max_sie_buffers + 1
    end
    local sid = self.stream_server:add_channel(stanza, cn.pipe_list)
    self.sid_list[stanza.name] = sid

    if set_time then
        self.set_time = self.set_time or {}
        self.set_time[#self.set_time + 1] = set_time
    end

    self:init_rezero(cn, stanza)
end

function m:initialize_generic_source(stanza)
    stanza.rezero_offset = self:preset_rezero(stanza)
    -- If there is a rezero offset then make sure the proxy reader doesn't try
    -- to use the scaling/offset readers (unless normal scale/offset is not
    -- 1.0/0.0).
    if stanza.rezero_offset then stanza.do_prerun_zero = nil end

    local pn, start, set_time = self:create_proxy_node(stanza)
    self.set_time = self.set_time or {}
    self.set_time[#self.set_time + 1] = set_time
    self.start_fn[#self.start_fn + 1] = {
        start = start,
        md = stanza,
    }
    self.proxy_nodes[stanza.connector] = pn
    self.all_nodes[stanza.name] = pn
    self.metas[stanza.name] = stanza
    if stanza.collect then
        pn.sie_idx = self.max_sie_buffers + SIE_IDX_OFFSET
        self.max_sie_buffers = self.max_sie_buffers + 1
    end

    -- KLUDGE:  If this is a sequentialized numeric message channel,
    -- change the data_type for it.  Note, this can not be done in
    -- the setup prepocessing function!
    local st = stanza.data_type:match('^message_(float)%d+')
    if st and stanza.sample_rate and
        tonumber(stanza.sample_rate) or 0 > 0 then
        stanza.data_type = string.gsub(stanza.data_type, 'message_',
            'sequential_')
        stanza.data_type = string.gsub(stanza.data_type, '_ntp$', '')
    end

    local sid = self.stream_server:add_channel(stanza, pn.pipe_list)
    self.sid_list[stanza.name] = sid

    self:init_rezero(pn, stanza)
end

function m:initialize_digital_output(stanza, pipe)
    -- Instantiate the proxy sink
    local pn = self:create_digital_out_node(stanza)

    self.module_nodes[#self.module_nodes + 1] = pn
    self.all_nodes[stanza.name] = pn
    self.metas[stanza.name] = stanza

    -- Attach the pipe from the input channel to this node
    pn.in_pipe = pipe.attach_pipe(
        self.all_nodes[stanza.input_channels[1]].pipe_list)
end

function m:initialize_mx878(stanza)
    local pn, start = self:create_mx878_proxy_node(stanza,
        self.setup[1].pipe_rate or 5)

    self.start_fn[#self.start_fn + 1] = {
        start = start,
        md = stanza,
    }
    self.proxy_nodes[stanza.connector] = pn
    self.all_nodes[stanza.name] = pn
    self.metas[stanza.name] = stanza
end

function m:_init_filter_sink_common(stanza)
    if not module_cache[stanza.category] then
        local ok, mod = pcall(require, 'test_engine/modules/'..stanza.category)
        if not ok then
            error(string.format('Unknown module type %s (%s)',
                stanza.category, stanza.name))
        end
        module_cache[stanza.category] = mod
    end
    return module_cache[stanza.category]
end

function m:initialize_filter(stanza)
    local m = self:_init_filter_sink_common(stanza)

    -- Instantiate the module.  Note that the module is responsible for
    -- Creating and attaching pipes to input nodes.  This also assumes
    -- the setup is sorted such that all channels are declared before
    -- being referenced!
    local new_mod, sie_buffers_needed = m:new(stanza, self.all_nodes,
        self.max_sie_buffers + SIE_IDX_OFFSET, self.metas)
    if not new_mod then
        error(string.format('Unable to create type %s (%s)',
            stanza.category, stanza.name))
    end
    if not sie_buffers_needed then
        sie_buffers_needed = new_mod.sie_buffers_needed
        if not sie_buffers_needed then
            error(string.format('Unable get SIE buf count for %s (%s)',
                stanza.category, stanza.name))
        end
    end

    -- KLUDGE workaround for filter module's returning incorrect values
    -- for the number of SIE buffers needed when the channel is NOT
    -- being collected.  In that case, the filter does NOT need any
    -- SIE buffers and none will have been allocated for it so override
    -- the sie_buffers_needed that was returned!
    if not stanza.collect then sie_buffers_needed = 0 end

    self.module_nodes[#self.module_nodes + 1] = new_mod
    self.all_nodes[stanza.name] = new_mod
    self.metas[stanza.name] = stanza
    self.max_sie_buffers = self.max_sie_buffers + sie_buffers_needed
    local sid = self.stream_server:add_channel(stanza, new_mod.pipe_list)
    self.sid_list[stanza.name] = sid
end

function m:initialize_sink_sie(stanza)
    local m = self:_init_filter_sink_common(stanza)

    -- Instantiate the module.  Note that the module is responsible for
    -- Creating and attaching pipes to input nodes.  This also assumes
    -- the setup is sorted such that all channels are declared before
    -- being referenced!
    local new_mod, sie_buffers_needed = m:new(stanza, self.all_nodes,
        self.max_sie_buffers + SIE_IDX_OFFSET, self.metas)
    if not new_mod then
        error(string.format('Unable to create type %s (%s)',
            stanza.category, stanza.name))
    end
    if not sie_buffers_needed then
        sie_buffers_needed = new_mod.sie_buffers_needed
        if not sie_buffers_needed then
            error(string.format('Unable get SIE buf count for %s (%s)',
                stanza.category, stanza.name))
        end
    end

    self.module_nodes[#self.module_nodes + 1] = new_mod
    self.all_nodes[stanza.name] = new_mod
    self.metas[stanza.name] = stanza
    self.max_sie_buffers = self.max_sie_buffers + sie_buffers_needed
end

function m:initialize_non_local_node(stanza, pipe)
    -- Channels in a mode 2 networked test that are NOT local to
    -- this node simply get added to the stream server.  Sink channels
    -- are NEVER added to the stream server.
    if not stanza.category:find('^sink/') then
        -- Add the channel to the stream server regardless.  In this
        -- case, a subscribe to the channel results in a connection to
        -- the appropriate slave's stream server!
        -- GPS channels other than the "pps_offset" and "timestamp"
        -- channels have already been added at this point.
        local sid
        if not stanza.category:find('^source/') then
            -- Computed channels are on the stream
            sid = self.stream_server:add_channel(stanza)
        elseif stanza.connector and stanza.data_type and
            stanza.data_type:find('_float%d+') then
            -- All numerice data type source channels are on the stream
            sid = self.stream_server:add_channel(stanza)
        end
        if sid then self.sid_list[stanza.name] = sid end
    elseif stanza.category:find('^sink/.*/di.+output$') then
-- TODO: Figure this out in the preprocessor
        -- Digital outputs are always put on the node from which the
        -- input channel resides
        local function input_channel_is_local(name)
            for _, s in ipairs(self.setup) do
                if s.name and s.name == name and s.is_local then
                    return true
                end
            end
        end

        if input_channel_is_local(stanza.input_channels[1]) then
            -- Instantiate the proxy sink
            local pn = self:create_digital_out_node(stanza)

            self.module_nodes[#self.module_nodes + 1] = pn
            self.all_nodes[stanza.name] = pn
            self.metas[stanza.name] = stanza

            -- Attach the pipe from the input channel to this node
            pn.in_pipe = pipe.attach_pipe(
                self.all_nodes[stanza.input_channels[1]].pipe_list)
        end
    end
end

function m:initialize_local_node(stanza, pipe)
    -- If we are NOT mode 2 networking OR we are and the channel is local
    -- create the appropriate proxy nodes for each channel type AND add
    -- the channel to the HBM stream server.
    if stanza.category and stanza.category:find('msg_can$') then
        self:initialize_raw_can(stanza)

    elseif stanza.category and stanza.category:find('can$') then
        self:initialize_can_signal(stanza)

    elseif stanza.category and stanza.category:find('msg_gps$') then
        if stanza.connector and stanza.connector:find('utc_date_time') then
            self:initialize_gps_signal(stanza)
        else
            self:initialize_raw_gps(stanza)
        end

    -- Note: GPS pps_offset and timestamp channels are handled as common
    -- "source" channels
    elseif stanza.category and stanza.category:find('gps$') and
        not stanza.connector:find('pps_offset') and
        not stanza.connector:find('timestamp') then
        self:initialize_gps_signal(stanza)

    elseif stanza.category and stanza.category:find('mx471') then
-- TODO: Need to test this channel type
        self:initialize_mx471(stanza)

    elseif stanza.category and stanza.category:find('^source') then
        self:initialize_generic_source(stanza)

    elseif stanza.category and (stanza.category:find('^sink/hbm_cx23r/') or
        stanza.category:find('^sink/hbm_edaqxr.*/')) then
        self:initialize_digital_output(stanza, pipe)

    elseif stanza.category and
        stanza.category:find('^sink/hbm_mx878b$') then
-- TODO: Need to test this channel type
        self:initialize_mx878(stanza)

    elseif stanza.category and stanza.category:find('^filter') then
        self:initialize_filter(stanza)

    elseif stanza.category and stanza.category:find('^sink/sie') then
        self:initialize_sink_sie(stanza)
    end
end

function m:initialize_node(stanza, hostname, pipe)
    if self.is_mode2_networking and stanza.network_node ~= hostname and
        not stanza.is_local then
        self:initialize_non_local_node(stanza, pipe)
    else
        self:initialize_local_node(stanza, pipe)
    end
end

return m
