-- Miscellaneous functions used in the test engine
local network_db = require 'network_db'
local hbm_stream_server = require 'test_engine.hbm_stream_server'
local ev = require 'ev'


local m = {}

-- Function used to create an HBM stream server and generate the configuration
-- information needed by the NGINX server to serve as a proxy for it.
function m:config_nginx_stream(hash)
    self.stream_server =
        hbm_stream_server:new(ev.Loop.default, {port = 8080, cport = 8081})
end

-- Function use to determine if all source channels in a test are available
function m:all_channels_available()
    -- Get the list of physical connectors from the setup
    local net_db = network_db:new()
    for _, stanza in pairs(self.setup) do
        local connector = stanza.connector
        if connector then
            if connector:find('@can') then
                connector = string.gsub(connector, '.+@', '@')
            end

            -- If the connector can not be resolved, the channel is not
            -- available.  So, throw an error
            if not net_db:resolve_service(connector:match('.*@(.+)')) then
                error(string.format("Channel %s is not available", connector))
            end
        end
    end
    net_db:close()
end

function m:_get_upstream_datatype(chans)
    local dt
    -- Check each of the input channels
    for _, name in ipairs(chans) do
        -- Find the input channel stanza in the setup
        local st
        for _, sn in ipairs(self.setup) do
            if sn.name == name then
                st = sn
                break
            end
        end
        if not st then
            error(string.format('%s:  no such channel in test', name))
        end
        -- Throw error if the input channel does NOT have a datatype.
        if not st.data_type then
            error(string.format('%s:  missing datatype', st.name))
        end
        -- If the data_type has not been set, set it.  If it
        -- has, make sure it is the same as the data type for
        -- this input channel.  Throw error if not
        dt = dt or st.data_type
        if dt ~= st.data_type then
            error(string.format(
                '%s:  inconsistant input channel data types', st.name))
        end
    end
    return dt
end

return m
